/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/
/**
 * @file CommunicationManagerImpl.h
 *
 * @date 2011-10-26
 * @author Philipp Kegel
 */

#ifndef COMMUNICATIONMANAGERIMPL_H_
#define COMMUNICATIONMANAGERIMPL_H_

#include <dcl/CommunicationManager.h>

#include <gcf/CommunicationManager.h>
#include <gcf/CommunicationManagerProperties.h>
#include <gcf/ConnectionListener.h>
#include <gcf/GCFTypedefs.h>
#include <gcf/MessageListener.h>
#include <gcf/Process.h>
#include <gcf/ReplicateMappings.h>
#include <gcf/Serializable.h>
#include <gcf/StreamCommunication.h>

#include <gcf/transport/DataStream.h>
#include <gcf/transport/IODataStream.h>

#include <chrono>
#include <map>
#include <memory>
#include <mutex>
#include <string>
#include <vector>

namespace dclgcf {

/* forward declarations */
class ComputeNodeImpl;
class ProcessImpl;

/******************************************************************************/

/**
 * @brief An abstract base class for managing communication.
 *
 * Override methods {host|resourceManager|computeNode}Connection{Established|Lost|Approved}
 * to define behavior in derived class.
 */
class CommunicationManagerImpl:
        public virtual dcl::CommunicationManager, // implement CommunicationManager
        public gcf::ConnectionListener,
        public gcf::DataStreamListener,
        public gcf::MessageListener {
public:
    /** default connection timeout is 3 seconds */
    static const std::chrono::seconds DEFAULT_CONNECTION_TIMEOUT;

    static const gcf_uint16 DEFAULT_COMPUTE_NODE_PORT = 25025;

    /**
     * @brief Extracts host name and port from a string.
     *
     * @param[in]     url       the URL to resolve
     * @param[out]    hostName  a host name
     * @param[in,out] port      default port.
     * If the URL contains a port number, it is returned by this parameter.
     */
    static void resolveURL(
            const std::string&  url,
            std::string&        hostName,
            gcf_uint16&         port);


    CommunicationManagerImpl(
            const gcf::CommunicationManagerProperties& properties);
    virtual ~CommunicationManagerImpl();

    void start();
    void stop();

    /**
     * @brief Returns the dOpenCL process that is associated with a specified GCF process.
     *
     * Subclasses may override this method in order to introduce additional
     * process types.
     *
     * @param[in]  process  a GCF process
     * @return the dOpenCL process that is associated with the specified GCF
     *         process, or \c NULL if no dOpenCL process is associated with the GCF
     *         process.
     */
    virtual ProcessImpl * getProcess(
            const gcf::Process& process) const;

    ComputeNodeImpl * getComputeNode(
            const gcf::Process& process) const;

    /*
     * GCF connection listener API
     */
    bool approveConnection(
            const std::string& subjectName);
    void connectionEstablished(
            const gcf::Process& process,
            const std::string&  subjectName);
    void connectionLost(
            const gcf::Process& process);
    gcf::ReplicateMappings incomingConnectionTransfer(
            const gcf::Process&                 host,
            const gcf::Process&                 computeNode,
            gcf::Serializable *                 transferMessage,
            std::vector<gcf::Serializable *>&   connectedEntities,
            std::vector<gcf::Serializable *>&   newEntities);

    /*
     * GCF data stream listener API
     */
    bool incomingDataStream(
            gcf::IODataStream& dataStream);
    void closedDataStream(
            gcf::DataStream& dataStream);

    /*
     * GCF message listener API
     */
    void dispatch(
            gcf::Serializable&  message,
            const gcf::Process& process);

protected:
    /* Connection managers must be non-copyable */
    CommunicationManagerImpl(
            const CommunicationManagerImpl&) = delete;
    CommunicationManagerImpl& operator=(
            const CommunicationManagerImpl&) = delete;

    /**
     * @brief Processes an incoming host connection.
     * A connection approval is send to the remove process if it has been
     * associated with a host instance successfully.
     *
     * @param[in]  process  the process to associate with a new host instance
     */
    virtual void hostConnectionEstablished(
            const gcf::Process& process);
    virtual void computeNodeConnectionEstablished(
            const gcf::Process& process);

    /**
     * @brief Tests, if a lost connection is associated with a host.
     *
     * @param[in]  process  process which lost connection
     * @return true, if @c process is associated with a host, otherwise false.
     */
    virtual bool hostConnectionLost(
            const gcf::Process& process);
    virtual bool computeNodeConnectionLost(
            const gcf::Process& process);

    /**
     * @brief Processes an incoming connection approval.
     *
     * @param[in]  process  the process which sent the connection approval
     */
    void connectionApproved(
            const gcf::Process& process) const;

    std::unique_ptr<gcf::CommunicationManager> _communicationManager;

    /* TODO Use process ID as comparator for map of compute nodes */
    /* TODO Use std::map<const gcf::Process *, std::unique_ptr<ComputeNodeImpl>> to store a communication manager's compute nodes */
    std::map<const gcf::Process *, ComputeNodeImpl *> _computeNodes;
    mutable std::recursive_mutex _connectionsMutex;
};

} /* namespace dclgcf */

#endif /* COMMUNICATIONMANAGERIMPL_H_ */
