/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/
/**
 * @file ComputeNodeCommunicationManagerImpl.h
 *
 * @date 2011-10-26
 * @author Philipp Kegel
 */

#ifndef COMPUTENODECOMMUNICATIONMANAGERIMPL_H_
#define COMPUTENODECOMMUNICATIONMANAGERIMPL_H_

#include "CommunicationManagerImpl.h"
#include "SmartCLObjectRegistry.h"

#include <dcl/CommunicationManager.h>
#include <dcl/ComputeNode.h>
#include <dcl/ConnectionListener.h>
#include <dcl/Daemon.h>

#include <gcf/CommunicationManager.h>
#include <gcf/CommunicationManagerProperties.h>
#include <gcf/ConnectionListener.h>
#include <gcf/MessageListener.h>
#include <gcf/Process.h>
#include <gcf/ReplicateMappings.h>
#include <gcf/Serializable.h>

#include <map>
#include <memory>
#include <mutex>
#include <set>
#include <string>
#include <utility>
#include <vector>

namespace dclgcf {

/* forward declarations */
class ComputeNodeImpl;
class HostImpl;
class ProcessImpl;

namespace comm {

class CLHostMessageDispatcher;
class CLRequestDispatcher;

} /* namespace comm */

/******************************************************************************/

class ComputeNodeCommunicationManagerImpl:
        public CommunicationManagerImpl,              // extend CommunicationManagerImpl
        public dcl::ComputeNodeCommunicationManager { // implement ComputeNodeCommunicationManager
public:
    ComputeNodeCommunicationManagerImpl(
            const gcf::CommunicationManagerProperties& properties);
    virtual ~ComputeNodeCommunicationManagerImpl();

    void start();
    void stop();

    /**
     * @brief Returns this communication manager's application object registry.
     *
     * @return an application object registry
     */
    SmartCLObjectRegistry& objectRegistry();

    void createComputeNodes(
            const std::vector<std::string>& urls,
            std::vector<ComputeNodeImpl *>& computeNodes);
    /**
     * @brief Creates compute nodes from the given GCF processes.
     *
     * WARNING: Compute nodes returned by this method may not be connected such
     * that subsequent communication with these compute nodes will fail.
     */
    void createComputeNodes(
            const std::set<const gcf::Process *>&   processes,
            std::vector<ComputeNodeImpl *>&         computeNodes);
    void destroyComputeNode(
            ComputeNodeImpl *computeNode);

    void connectComputeNodes(
            const std::vector<ComputeNodeImpl *>& computeNodes);

    void setDaemon(
            dcl::Daemon *daemon = nullptr);
    dcl::Daemon * getDaemon() const;

    bool addConnectionListener(
            dcl::ConnectionListener& listener);
    bool removeConnectionListener(
            dcl::ConnectionListener& listener);

    /**
     * @brief Returns the dOpenCL process that is associated with a specified GCF process.
     *
     * This method overrides CommunicationManagerImpl::getProcess in order to also
     * search the list of hosts.
     */
    ProcessImpl * getProcess(
            const gcf::Process& process) const;

    HostImpl * getHost(
            const gcf::Process& process) const;

protected:
    void hostConnectionEstablished(
            const gcf::Process& process);
    void computeNodeConnectionEstablished(
            const gcf::Process& process);

    bool hostConnectionLost(
            const gcf::Process& process);
    bool computeNodeConnectionLost(
            const gcf::Process& process);

private:
    SmartCLObjectRegistry _objectRegistry; /**< Registry for application objects */

    std::unique_ptr<comm::CLHostMessageDispatcher> _clMessageDispatcher;
    std::unique_ptr<comm::CLRequestDispatcher> _clRequestDispatcher;

    dcl::Daemon *_daemon;
    std::set<dcl::ConnectionListener *> _connectionListeners;
    std::mutex _connectionListenersMutex;

    /* TODO Use std::map<const gcf::Process *, std::unique_ptr<HostImpl>> */
    std::map<const gcf::Process *, HostImpl *> _hosts;
};

} /* namespace dclgcf */

#endif /* COMPUTENODECOMMUNICATIONMANAGERIMPL_H_ */
