/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/
/*!
 * \file ComputeNodeImpl.h
 *
 * \date 2011-10-30
 * \author Philipp Kegel
 */

#ifndef COMPUTENODEIMPL_H_
#define COMPUTENODEIMPL_H_

#include "ProcessImpl.h"

#include "comm/ResponseBuffer.h"

#include "message/ProcessMessage.h"

#include <dclgcf/message/Request.h>
#include <dclgcf/message/Response.h>

#include <dcl/Binary.h>
#include <dcl/ComputeNode.h>
#include <dcl/DCLException.h>
#include <dcl/Device.h>

#include <dcl/util/Logger.h>

#include <gcf/CommunicationManager.h>
#include <gcf/GCFTypedefs.h>
#include <gcf/Process.h>
#include <gcf/StreamCommunication.h>

#include <gcf/link/SocketOptions.h>

#ifdef __APPLE__
#include <OpenCL/cl_wwu_dcl.h>
#else
#include <CL/cl_wwu_dcl.h>
#endif

#include <chrono>
#include <memory>
#include <mutex>
#include <vector>

namespace dclgcf {

/* forward declarations */
class DeviceImpl;

/* ****************************************************************************/

class ComputeNodeImpl:
public dcl::ComputeNode, // implements ComputeNode
public ProcessImpl {     // extends ProcessImpl
public:
    /*!
     * \brief Connects to multiple compute nodes concurrently.
     *
     * This method updates the compute nodes' device lists.
     * This method works on a best-effort base: failed connection attempts are
     * ignored.
     *
     * \param[in]  computeNodes     the compute nodes to connect to
     * \param[in]  localProcessType type of the local process that is establishing the connection
     */
    static void connect(
            const std::vector<ComputeNodeImpl *>&   computeNodes,
            Type                                    localProcessType);

    /*!
     * \brief Awaits for multiple compute nodes to connect.
     *
     * \param[in]  computeNodes the compute nodes which connection to wait for
     */
    static void awaitConnection(
            const std::vector<ComputeNodeImpl *>& computeNodes);


    /*!
     * \brief Creates a compute node instance.
     *
     * \param[in]  communicationManager an associated GCF communication manager
     * \param[in]  process              an associated GCF process
     */
    ComputeNodeImpl(
            gcf::CommunicationManager& communicationManager,
            const gcf::Process&        process);
    virtual ~ComputeNodeImpl();

    /*!
     * \brief Connects to the compute node
     *
     * A connection is established in four steps:
     * 1. connect GCF process
     * 2. await session creation on compute node
     * 3. create GCF I/O data stream
     * 4. update device list
     *
     * \param[in]  localProcessType type of the local process that is establishing the connection
     */
    /* TODO Allow for asynchronous connect using a callback */
    void connect(
            Type localProcessType);

    void getDevices(
            std::vector<dcl::Device *>& devices);

    void getInfo(
            cl_compute_node_info_WWU    param_name,
            dcl::Binary&                param) const;

    void sendRequest(
            message::Request& request) const;

    std::unique_ptr<message::Response> awaitResponse(
            const message::Request& request,
            GCFTypeID_t             responseType);
    void awaitResponse(
            const message::Request& request);

    std::unique_ptr<message::Response> executeCommand(
            message::Request&   request,
            GCFTypeID_t         responseType);
    void executeCommand(
            message::Request& request);

    comm::ResponseBuffer& responseBuffer();

private:
    /*!
     * \brief Updates the device lists of multiple compute nodes.
     *
     * The device IDs are queried from multiple remote compute nodes.
     * This method is more efficient than @c updateDevices.
     *
     * \param[in]  computeNode  the compute node to query the device IDs from
     */
    static void updateDevices(
            const std::vector<ComputeNodeImpl *>& computeNodes);

    /*!
     * \brief Connects to the compute node's GCF process.
     *
     * \param[in]  localProcessType type of the local process that is establishing the connection
     * \param[in]  timeout          maximum duration of connection process
     */
    template<class Rep, class Period>
    void connectProcess(
            Type localProcessType,
            std::chrono::duration<Rep, Period> timeout) {
        gcf::SocketOptions socketOptions;

        /* Disable Nagle's algorithm
         * Due to the RPC-style protocol of dOpenCL, short messages usually wait for
         * a response before the next message is send. Hence, waiting for another
         * short message to merge with the first one is pointless.
         * In particular, command status messages suffer from the high latency
         * introduced by Nagle's algorithm. */
        socketOptions.set("SOL_TCP", "TCP_NODELAY");
        /* CommunicationManager::setOptions set the options locally and also sends a
         * message to the remote process in order to let that process set the same
         * options. However, this does not work reliably (at least for TCP_NODELAY).
         * Moreover, the options are sent twice (in CommunicationManager::setOptions
         * and CommunicationManager::setOptions); maybe this is a GCF bug */

        try {
            /*
             * connect process
             */
            {
                std::lock_guard<std::recursive_mutex> lock(_connectionStatusMutex);
                bool connected = _communicationManager.preconnectProcess(
                        _process,
                        std::chrono::duration_cast<std::chrono::milliseconds>(timeout).count(),
                        socketOptions);
                if (!connected) { /* timeout reached or failure */
                    throw dcl::ConnectionException(
                            "Could not connect to compute node at " + url());
                }
                _connectionStatus = ConnectionStatus::CONNECTING;
                _connectionStatusChanged.notify_all();
            }

            dcl::util::Logger << dcl::util::Info
                    << "Connecting to compute node " << url() << std::endl;

            /* Inform remote process of the type of dOpenCL process that has
             * established the incoming connection */
            switch (localProcessType) {
            case Type::HOST:
            {
                message::HostIdentificationMessage message;
                _communicationManager.send(_process, message);
                break;
            }
            case Type::COMPUTE_NODE:
            {
                message::ComputeNodeIdentificationMessage message;
                _communicationManager.send(_process, message);
                break;
            }
            default:
                /* TODO Throw exception
                 * Processes other than hosts or compute nodes must not create
                 * compute node connections */
                ;
                /* no break */
            }
        } catch (const gcf::GCFError& err) {
            throw dcl::ConnectionException(err.message);
        }
    }

    /*!
     * \brief Initializes the compute node's GCF I/O data stream.
     *
     * \param[in]  timeout  a timeout; may be a duration or time point
     */
    template<typename timeout_type>
    void createDataStream(
            const timeout_type& timeout) {
        /*
         * Await connection approval
         * The peer's listening socket for data streams must have been initialized
         * before a data stream can be created.
         */
        bool connected = awaitConnectionStatus(ConnectionStatus::CONNECTED, timeout);
        if (!connected) { /* session creation timed out */
            throw dcl::ConnectionException("Could not connect to compute node at " + url());
        }

        /*
         * create I/O data stream
         */
        {
            std::lock_guard<std::recursive_mutex> lock(_connectionStatusMutex);
            gcf::StreamCommunication& sc = _communicationManager.getStreamCommunication();
            _dataStream = &sc.createIODataStream(_process);
        }
    }

    /*!
     * \brief Updates the compute nodes device list.
     * The device IDs are queried from the remote compute node.
     */
    void updateDevices();
    /*!
     * \brief Updates the compute node's device list using a list of device IDs.
     *
     * \param[in]  deviceIds    the device IDs that should be used for the update
     */
    void updateDevices(
            const std::vector<dcl::object_id>& deviceIds);


    comm::ResponseBuffer _responseBuffer; //!< Response buffer

    std::unique_ptr<std::vector<std::unique_ptr<DeviceImpl>>> _devices; //!< Device list
    /*!
     * \brief A mutex associated with this compute node's devices list.
     *
     * This mutex must be recursive as getDeviceIDs locks it and will attempt to
     * lock it again when the device IDs have to be queried in updateDevices.
     */
    std::recursive_mutex _devicesMutex;
};

/* ****************************************************************************/

/*!
 * \brief Sends a request to multiple compute nodes.
 *
 * \param[in]  computeNodes a list of compute nodes to send a request to
 * \param[in]  request      the request to send
 */
void sendRequest(
        const std::vector<ComputeNodeImpl *>&   computeNodes,
        message::Request&                       request);

/*!
 * \brief Executes a command on multiple compute nodes.
 *
 * \param[in]  computeNodes a list of compute node to execute the command on
 * \param[in]  request      the request for a given command
 * \param[in]  responseType the expected type of responses
 * \param[out] responses    the responses returned by the compute nodes
 */
void executeCommand(
        const std::vector<ComputeNodeImpl *>&               computeNodes,
        message::Request&                                   request,
        GCFTypeID_t                                         responseType = message::Response::TYPE,
        std::vector<std::unique_ptr<message::Response>> *   responses = nullptr);

} /* namespace dclgcf */

#endif /* COMPUTENODEIMPL_H_ */
