/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/
/**
 * @file CLMessageDispatcher.cpp
 *
 * @date 2011-12-16
 * @author Philipp Kegel
 * @author Sebastian Pribnow
 */

#include "CLMessageDispatcher.h"

#include "../CommunicationManagerImpl.h"
#include "../ComputeNodeCommunicationManagerImpl.h"
#include "../ComputeNodeImpl.h"
#include "../HostImpl.h"
#include "../SmartCLObjectRegistry.h"

#include <dclgcf/message/CommandMessage.h>
#include "../message/ContextErrorMessage.h"
#include <dclgcf/message/EventSynchronizationMessage.h>
#include "../message/ProgramBuildMessage.h"

#include <dcl/Binary.h>
#include <dcl/BlockingQueue.h>
#include <dcl/CLObjectRegistry.h>
#include <dcl/CommandListener.h>
#include <dcl/ContextListener.h>
#include <dcl/DCLException.h>
#include <dcl/DCLTypes.h>
#include <dcl/Device.h>
#include <dcl/Process.h>
#include <dcl/ProgramBuildListener.h>
#include <dcl/SynchronizationListener.h>

#include <dcl/util/Logger.h>

#include <gcf/CallbackPolicy.h>
#include <gcf/GCFTypedefs.h>

#ifdef __APPLE__
#include <OpenCL/cl.h>
#else
#include <CL/cl.h>
#endif

#include <functional>
#include <map>
#include <memory>
#include <ostream>
#include <string>
#include <thread>
#include <vector>

namespace dclgcf {

namespace comm {

CLMessageDispatcher::MessageTypes::MessageTypes() {
    /* add context message types */
    push_back(message::ContextErrorMessage::TYPE);
    /* add event message types */
    push_back(message::CommandExecutionStatusChangedMessage::TYPE);
    /* add program build message types */
    push_back(message::ProgramBuildMessage::TYPE);
    /* add memory consistency message types */
    push_back(message::EventSynchronizationMessage::TYPE);
}

/******************************************************************************/

const gcf::CallbackPolicy& CLMessageDispatcher::callbackPolicy() {
    static gcf::CallbackPolicy policy(
            MessageTypes(),
            false /* do not exclude but include specified types */);
    return policy;
}

/******************************************************************************
 * Host side message dispatcher
 ******************************************************************************/

CLComputeNodeMessageDispatcher::CLComputeNodeMessageDispatcher(
        const CommunicationManagerImpl& connectionManager,
        const dcl::CLObjectRegistry& objectRegistry) :
    _communicationManager(connectionManager), _objectRegistry(objectRegistry) {
    /* Message dispatcher must be initialized before the worker thread is created */
    _worker = std::thread(&CLComputeNodeMessageDispatcher::run, this);
}

CLComputeNodeMessageDispatcher::~CLComputeNodeMessageDispatcher() {
    /* interrupt blocking queue to stop thread */
    _taskList.interrupt();
    if (_worker.joinable()) _worker.join();
}

void CLComputeNodeMessageDispatcher::run() {
    try {
        /* wait for and dispatch execution status changes */
        while (true) {
            _taskList.front()();
            _taskList.pop();
        }
    } catch (const dcl::ThreadInterrupted&) {
        /* task queue has been interrupted */
    }
}

void CLComputeNodeMessageDispatcher::contextError(
        const message::ContextErrorMessage& notification) const {
    dcl::ContextListener *contextListener = _objectRegistry.lookup<dcl::ContextListener>(notification.contextId());
    if (contextListener) {
        contextListener->onError(notification.errorInfo().c_str(),
                notification.privateInfo().value(), notification.privateInfo().size());
    } else {
        dcl::util::Logger << dcl::util::Error
                << "Context listener not found (ID=" << notification.contextId()
                << ')' << std::endl;
    }
}

void CLComputeNodeMessageDispatcher::executionStatusChanged(
        const message::CommandExecutionStatusChangedMessage& notification) {
    dcl::CommandListener *commandListener = _objectRegistry.lookup<dcl::CommandListener>(notification.commandId());
    if (commandListener) {
        /* pass function call to worker thread */
        _taskList.push(
                std::bind(&dcl::CommandListener::onExecutionStatusChanged,
                        commandListener, notification.status()));
    } else {
        dcl::util::Logger << dcl::util::Error
                << "Command listener not found (command ID=" << notification.commandId()
                << ')' << std::endl;
    }
}

void CLComputeNodeMessageDispatcher::synchronizeEvent(
        const message::EventSynchronizationMessage& notification,
        dcl::Process& process) const {
    dcl::SynchronizationListener *synchronizationlistener = _objectRegistry.lookup<dcl::SynchronizationListener>(notification.commandId());
    if (synchronizationlistener) {
        synchronizationlistener->onSynchronize(process);
    } else {
        dcl::util::Logger << dcl::util::Error
                << "Synchronization listener not found (command ID=" << notification.commandId()
                << ')' << std::endl;
    }
}

void CLComputeNodeMessageDispatcher::programBuildComplete(
        const message::ProgramBuildMessage& notification) const {
    dcl::ProgramBuildListener *programBuildListener = _objectRegistry.lookup<dcl::ProgramBuildListener>(notification.programBuildId());
    if (programBuildListener) {
        std::vector<dcl::Device *> devices;
        std::vector<cl_build_status> buildStatus;

        /* TODO Lookup devices from notification.deviceIds() */

        notification.getBuildStatus(buildStatus);
        programBuildListener->onComplete(devices, buildStatus);
    } else {
        dcl::util::Logger << dcl::util::Error
                << "Program build listener not found (ID=" << notification.programBuildId()
                << ')' << std::endl;
    }
}

/*
 * GCF message listener API
 */

void CLComputeNodeMessageDispatcher::dispatch(gcf::Serializable& message,
        const gcf::Process& process) {
    ComputeNodeImpl *computeNode = _communicationManager.getComputeNode(process);
    assert(computeNode != nullptr);

    switch (message.getType()) {

    /*
     * Asynchronous notification messages
     * Messages must be copied as they will be deleted by the GCF when dispatch
     * returns.
     */
    case message::ContextErrorMessage::TYPE:
        dcl::util::Logger << dcl::util::Debug
                << "Received context error message from compute node" << std::endl;
        contextError(
                static_cast<message::ContextErrorMessage&>(message));
        break;

    case message::CommandExecutionStatusChangedMessage::TYPE:
        dcl::util::Logger << dcl::util::Debug
                << "Received command execution status changed message from compute node" << std::endl;
        executionStatusChanged(
                static_cast<message::CommandExecutionStatusChangedMessage&>(message));
        break;

    case message::EventSynchronizationMessage::TYPE:
        dcl::util::Logger << dcl::util::Debug
                << "Received event synchronization message from compute node" << std::endl;
        synchronizeEvent(
                static_cast<message::EventSynchronizationMessage&>(message),
                *computeNode);
        break;

    case message::ProgramBuildMessage::TYPE:
        dcl::util::Logger << dcl::util::Debug
                << "Received program build message" << std::endl;
        programBuildComplete(
                static_cast<message::ProgramBuildMessage&>(message));
        break;

    default: /* invalid message */
        dcl::util::Logger << dcl::util::Error
                << "Received unknown message" << std::endl;
        break;
    }
}

/******************************************************************************
 * Compute node side message dispatcher
 ******************************************************************************/

CLHostMessageDispatcher::CLHostMessageDispatcher(
        const ComputeNodeCommunicationManagerImpl& connectionManager) :
    _connectionManager(connectionManager) { }

SmartCLObjectRegistry& CLHostMessageDispatcher::getObjectRegistry(
        HostImpl& host) const {
    /* TODO Do not make registry a member of host */
    return host.objectRegistry();
}

void CLHostMessageDispatcher::executionStatusChanged(
        const message::CommandExecutionStatusChangedMessage& notification,
        HostImpl& host) const {
    auto event = getObjectRegistry(host).lookup<std::shared_ptr<dcl::Event>>(notification.commandId());
    if (event) {
        event->onExecutionStatusChanged(notification.status());
    } else {
        dcl::util::Logger << dcl::util::Error
                << "Event not found (command ID=" << notification.commandId()
                << ')' << std::endl;
    }
}

void CLHostMessageDispatcher::synchronizeEvent(
        const message::EventSynchronizationMessage& notification,
        HostImpl& host) const {
    auto event = getObjectRegistry(host).lookup<std::shared_ptr<dcl::Event>>(notification.commandId());
    if (event) {
        event->onSynchronize(host);
    } else {
        dcl::util::Logger << dcl::util::Error
                << "Event not found (command ID=" << notification.commandId()
                << ')' << std::endl;
    }
}

/*
 * GCF message listener API
 */

void CLHostMessageDispatcher::dispatch(gcf::Serializable& message,
        const gcf::Process& process) {
    HostImpl *host = _connectionManager.getHost(process);
    assert(host != nullptr);

    switch (message.getType()) {
    case message::CommandExecutionStatusChangedMessage::TYPE:
        dcl::util::Logger << dcl::util::Debug
                << "Received command execution status changed message from host" << std::endl;
        executionStatusChanged(
                static_cast<message::CommandExecutionStatusChangedMessage&>(message), *host);
        break;
    case message::EventSynchronizationMessage::TYPE:
        dcl::util::Logger << dcl::util::Debug
                << "Received event synchronization message from host" << std::endl;
        synchronizeEvent(
                static_cast<message::EventSynchronizationMessage&>(message), *host);
        break;

    default: /* invalid message */
        dcl::util::Logger << dcl::util::Error
                << "Received unknown message" << std::endl;
        break;
    }
}

} /* namespace comm */

} /* namespace dclgcf */
