/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/
/**
 * @file CLMessageDispatcher.h
 *
 * @date 2011-12-16
 * @author Philipp Kegel
 * @author Sebastian Pribnow
 */

#ifndef CLMESSAGEDISPATCHER_H_
#define CLMESSAGEDISPATCHER_H_

#include <dclgcf/message/CommandMessage.h>
#include <dclgcf/message/EventSynchronizationMessage.h>

#include <dcl/BlockingQueue.h>
#include <dcl/CLObjectRegistry.h>
#include <dcl/DCLTypes.h>
#include <dcl/Process.h>

#include <gcf/CallbackPolicy.h>
#include <gcf/GCFTypedefs.h>
#include <gcf/MessageListener.h>
#include <gcf/Process.h>
#include <gcf/Serializable.h>

#ifdef __APPLE__
#include <OpenCL/cl.h>
#else
#include <CL/cl.h>
#endif

#include <functional>
#include <thread>

namespace dclgcf {

/* forward declarations */
class ComputeNodeCommunicationManagerImpl;
class CommunicationManagerImpl;
class HostImpl;
class SmartCLObjectRegistry;

namespace message {

class ContextErrorMessage;
class ProgramBuildMessage;

} /* namespace message */

/******************************************************************************/

namespace comm {

/**
 * @brief A dispatch for incoming application level notification messages.
 */
class CLMessageDispatcher: public gcf::MessageListener {
public:
    /**
     * @brief Returns a GCF callback policy comprising the message types processed by a message dispatcher.
     *
     * @return a callback policy
     */
    static const gcf::CallbackPolicy& callbackPolicy();


	virtual ~CLMessageDispatcher() { }

protected:
    /**
     * @brief A list of dOpenCL message types.
     *
     * This class is used internally for static initialization of a GCF callback
     * policy.
     */
    class MessageTypes : public std::vector<GCFTypeID_t> {
    public:
        MessageTypes();
    };
};

/******************************************************************************/

/**
 * @brief A dispatch for incoming messages from compute nodes.
 *
 * This dispatcher is used on the host! It is called 'CLComputeNodeMessageDispatcher'
 * as it dispatches compute node messages, i.e., messages from compute nodes.
 *
 * This class should become the only implementation of a message dispatcher.
 * Due to implementation issues of ID generation (each host uses its own ID
 * range) it currently has to be distinguished from message dispatchers used on
 * the compute node side.
 *
 * @see CLHostMessageDispatcher
 */
class CLComputeNodeMessageDispatcher: public CLMessageDispatcher {
public:
    /* A task is a nullary function */
    typedef std::function<void ()> Task;

    CLComputeNodeMessageDispatcher(
            const CommunicationManagerImpl& connectionManager,
            const dcl::CLObjectRegistry&    objectregistry);
    virtual ~CLComputeNodeMessageDispatcher();

    void run();

    /*
     * GCF message listener API
     */
    void dispatch(
            gcf::Serializable&  message,
            const gcf::Process& process);

private:
    void contextError(
            const message::ContextErrorMessage& notification) const;

    void executionStatusChanged(
            const message::CommandExecutionStatusChangedMessage& notification);

    /**
     * @brief Callback for an incoming event synchronization request
     *
     * @param[in]  notification the synchronization request
     * @param[in]  process      the process that requested the synchronization
     */
    void synchronizeEvent(
            const message::EventSynchronizationMessage& notification,
            dcl::Process&                               process) const;

    void programBuildComplete(
            const message::ProgramBuildMessage& notification) const;

    const CommunicationManagerImpl& _communicationManager;
    const dcl::CLObjectRegistry& _objectRegistry; /**< Registry for application objects */

    dcl::BlockingQueue<Task> _taskList; /**< Task list for worker thread */

    std::thread _worker; /**< Worker thread */
};

/******************************************************************************/

/**
 * @brief A dispatch for incoming messages from hosts.
 *
 * This dispatcher is used on compute nodes! It is called 'CLHostMessageDispatcher'
 * as it dispatches host messages, i.e., messages from the host.
 *
 * Messages from hosts have to be dispatched differently as the ID within these
 * messages refers to a particular host and, thus, are not unique if multiple
 * hosts are connected to a compute nodes.
 * Hence, a different message dispatcher is required for each host that only
 * processes messages it received from this associated host.
 *
 * In future version of dOpenCL UUID could be used to assign globally unique IDs
 * to objects, such that a message dispatcher does not have to consider the
 * message source to resolve these IDs. Then, CLHostMessageDispatcher and
 * CLComputeNodeMessageDispatcher can be unified.
 */
class CLHostMessageDispatcher: public CLMessageDispatcher {
public:
    CLHostMessageDispatcher(
            const ComputeNodeCommunicationManagerImpl& connectionManager);

    /*
     * GCF message listener API
     */
    void dispatch(
            gcf::Serializable&  message,
            const gcf::Process& process);

private:
    SmartCLObjectRegistry& getObjectRegistry(
            HostImpl& host) const;

    void executionStatusChanged(
            const message::CommandExecutionStatusChangedMessage&    notification,
            HostImpl&                                               host) const;

    /**
     * @brief Callback for an incoming event synchronization request
     *
     * @param[in]  notification the synchronization request
     * @param[in]  host         the host that requested the synchronization
     */
    void synchronizeEvent(
            const message::EventSynchronizationMessage& notification,
            HostImpl&                                   host) const;

    const ComputeNodeCommunicationManagerImpl& _connectionManager;
};

} /* namespace comm */

} /* namespace dclgcf */

#endif /* CLMESSAGEDISPATCHER_H_ */
