/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/
/**
 * @file CLResponseDispatcher.cpp
 *
 * @date 2012-07-14
 * @author Philipp Kegel
 */

#include "CLResponseDispatcher.h"

#include "ResponseBuffer.h"

#include "../CommunicationManagerImpl.h"
#include "../ComputeNodeImpl.h"

#include "../message/DeviceIDsResponse.h"
#include "../message/DeviceInfosResponse.h"

#include <dclgcf/message/ErrorResponse.h>
#include <dclgcf/message/EventProfilingInfosResponse.h>
#include <dclgcf/message/InfoResponse.h>
#include <dclgcf/message/Response.h>

#include <dcl/util/Logger.h>

#include <gcf/CallbackPolicy.h>
#include <gcf/Process.h>
#include <gcf/Serializable.h>

#include <cassert>
#include <memory>
#include <ostream>
#include <utility>

namespace dclgcf {

namespace comm {

CLResponseDispatcher::MessageTypes::MessageTypes() {
    /* add command response message types */
    push_back(message::Response::TYPE);
    push_back(message::DeviceIDsResponse::TYPE);
    push_back(message::DeviceInfosResponse::TYPE);
    push_back(message::ErrorResponse::TYPE);
    push_back(message::EventProfilingInfosReponse::TYPE);
    push_back(message::InfoResponse::TYPE);
}

/******************************************************************************/

const gcf::CallbackPolicy& CLResponseDispatcher::callbackPolicy() {
    static gcf::CallbackPolicy policy(
            MessageTypes(),
            false /* do not exclude but include specified types */);
    return policy;
}

CLResponseDispatcher::CLResponseDispatcher(
        const CommunicationManagerImpl& connectionManager) :
        _communicationManager(connectionManager) {
}

CLResponseDispatcher::~CLResponseDispatcher() {
}

/******************************************************************************
 * GCF message listener API
 ******************************************************************************/

void CLResponseDispatcher::dispatch(gcf::Serializable& message,
		const gcf::Process& process) {
	ComputeNodeImpl *computeNode;
	std::unique_ptr<message::Response> response;

	/*
	 * Find response buffer
	 */
	computeNode = _communicationManager.getComputeNode(process);
	assert(computeNode != nullptr);

	switch (message.getType()) {

	/*
	 * Command response messages
	 * Messages must be copied as they will be deleted by the GCF when dispatch
	 * returns.
	 */
	/* TODO Implement move semantics for messages, such that responses will not
	 *      be deeply copied for buffering */
	case message::DeviceIDsResponse::TYPE:
		response.reset(new message::DeviceIDsResponse(
				static_cast<message::DeviceIDsResponse&>(message)));
		break;
	case message::DeviceInfosResponse::TYPE:
		response.reset(new message::DeviceInfosResponse(
				static_cast<message::DeviceInfosResponse&>(message)));
		break;
	case message::ErrorResponse::TYPE:
		response.reset(new message::ErrorResponse(
				static_cast<message::ErrorResponse&>(message)));
		break;
	case message::EventProfilingInfosReponse::TYPE:
		response.reset(new message::EventProfilingInfosReponse(
				static_cast<message::EventProfilingInfosReponse&>(message)));
		break;
	case message::InfoResponse::TYPE:
		response.reset(new message::InfoResponse(
				static_cast<message::InfoResponse&>(message)));
		break;
	case message::Response::TYPE:
		response.reset(new message::Response(
				static_cast<message::Response&>(message)));
		break;

	default: /* invalid message */
		dcl::util::Logger << dcl::util::Error
				<< "Received message of unknown type" << std::endl;
		break;
	}

	if (response) {
	    computeNode->responseBuffer().put(std::move(response));
	}
}

} /* namespace comm */

} /* namespace dclgcf */
