/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/
/**
 * @file DataDispatcher.cpp
 *
 * @date 2011-02-16
 * @author Philipp Kegel
 */

#include "DataDispatcher.h"

#include "DataTransferImpl.h"

#include <dcl/DCLException.h>

#include <gcf/transport/IODataStream.h>

#include <cstddef>
#include <memory>
#include <thread>

namespace dclgcf {

namespace comm {

DataDispatcher::DataDispatcher() {
    /* Data dispatcher must be initialized before a thread is created */
    _worker = std::thread(&DataDispatcher::run, this);
}

DataDispatcher::~DataDispatcher() {
    /* interrupt blocking queue to stop thread */
    _dataTransfers.interrupt();
    if (_worker.joinable()) _worker.join();
}

std::shared_ptr<DataTransferImpl> DataDispatcher::receiveData(
		gcf::IODataStream& dataStream, size_t size, void *ptr) {
	std::shared_ptr<DataReceipt> receipt(
	        std::make_shared<DataReceipt>(dataStream.recv(ptr, size)));
	_dataTransfers.push(receipt);
	return receipt;
}

std::shared_ptr<DataTransferImpl> DataDispatcher::sendData(
		gcf::IODataStream& dataStream, size_t size, const void *ptr) {
	std::shared_ptr<DataSending> sending(
	        std::make_shared<DataSending>(dataStream.send(ptr, size)));
	_dataTransfers.push(sending);
	return sending;
}

void DataDispatcher::run() {
	try {
		/* Wait for, fetch, and execute tasks from task queue */
		while (true) {
			std::shared_ptr<DataTransferImpl> dataTransfer = _dataTransfers.front();
			_dataTransfers.pop();
			dataTransfer->execute();
		}
	} catch (const dcl::ThreadInterrupted&) {
		/* task queue has been interrupted */
	}
}

} /* namespace comm */

} /* namespace dclgcf */
