/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/
/**
 * @file EnqueueNDRangeKernel.cpp
 *
 * @date 2011-07-13
 * @author Tunc Taylan Turunc
 */

#include "utility.h"

#include <dclgcf/message/EnqueueNDRangeKernel.h>
#include <dclgcf/message/Request.h>

#include <dcl/DCLTypes.h>

#include <gcf/GCFTypedefs.h>
#include <gcf/GenericSerializerImplNew.cpp>
#include <gcf/Serializable.h>

#include <gcf/serialization/ByteBuffer.h>

#include <cassert>
#include <cstddef>
#include <vector>

namespace gcf {

template<>
size_t readCustomSTR<dclgcf::message::EnqueueNDRangeKernel> (
		dclgcf::message::EnqueueNDRangeKernel *object,
		const gcf::GenericSerializeTemplateRecord& gstr,
		const char *buffer) {
	assert(gstr.getType() == gcf::ST_Custom1);

	gcf::ByteBuffer buf(buffer, gstr.getLength());
	std::vector<gcf_uint32>& ndRange =
			*gcf::GenericSerializeTemplateRecord::getAttribute<
					dclgcf::message::EnqueueNDRangeKernel, std::vector<gcf_uint32>,
					gcf::ST_Custom1>(object, gstr);

	ndRange.clear();
	buf >> ndRange;

	return buf.getLength();
}

template<>
size_t writeCustomSTR<dclgcf::message::EnqueueNDRangeKernel> (
		dclgcf::message::EnqueueNDRangeKernel *object,
		const gcf::GenericSerializeTemplateRecord& gstr,
		char *buffer) {
	assert(gstr.getType() == gcf::ST_Custom1);

	gcf::ByteBuffer buf(buffer, gstr.getLength());
	std::vector<gcf_uint32>& ndRange =
			*gcf::GenericSerializeTemplateRecord::getAttribute<
					dclgcf::message::EnqueueNDRangeKernel, std::vector<gcf_uint32>,
					gcf::ST_Custom1>(object, gstr);

	buf << ndRange;

	return buf.getLength();
}

} /* namespace gcf */

/******************************************************************************/

namespace dclgcf {

namespace message {

EnqueueNDRangeKernel::EnqueueNDRangeKernel() :
	_commandQueueId(0), _commandId(0), _kernelId(0), _event(false) {
	initSerializable();
}

EnqueueNDRangeKernel::EnqueueNDRangeKernel(
		dcl::object_id commandQueueId,
		dcl::object_id commandId,
		dcl::object_id kernelId,
		const std::vector<size_t>& offset,
		const std::vector<size_t>& global,
		const std::vector<size_t>& local,
		const std::vector<dcl::object_id> *eventIdWaitList,
		bool event) :
	_commandQueueId(commandQueueId), _commandId(commandId), _kernelId(kernelId),
			_offset(offset.size()), _global(global.size()),
			_local(local.size()), _event(event) {
	initSerializable();

	/* convert work dimensions from host to network representation */
	workDimensionHtoN(offset, _offset);
	workDimensionHtoN(global, _global);
	workDimensionHtoN(local, _local);

	if (eventIdWaitList) {
		_eventIdWaitList = *eventIdWaitList;
	}
}

EnqueueNDRangeKernel::EnqueueNDRangeKernel(
		const EnqueueNDRangeKernel& rhs) :
	Request(rhs), _commandQueueId(rhs._commandQueueId), _commandId(rhs._commandId),
			_kernelId(rhs._kernelId), _global(rhs._global), _local(rhs._local),
			_eventIdWaitList(rhs._eventIdWaitList), _event(rhs._event) {
	initSerializable();
}

dcl::object_id EnqueueNDRangeKernel::commandQueueId() const {
	return _commandQueueId;
}

dcl::object_id EnqueueNDRangeKernel::commandId() const {
	return _commandId;
}

dcl::object_id EnqueueNDRangeKernel::kernelId() const {
	return _kernelId;
}

void EnqueueNDRangeKernel::offset(std::vector<size_t>& offset) const {
	workDimensionNtoH(_offset, offset);
}

void EnqueueNDRangeKernel::global(std::vector<size_t>& global) const {
	workDimensionNtoH(_global, global);
}

void EnqueueNDRangeKernel::local(std::vector<size_t>& local) const {
	workDimensionNtoH(_local, local);
}

const std::vector<dcl::object_id>& EnqueueNDRangeKernel::eventIdWaitList() const {
	return _eventIdWaitList;
}

bool EnqueueNDRangeKernel::event() const {
	return static_cast<bool>(_event);
}

} /* namespace message */

} /* namespace dclgcf */

IMPLEMENT_SERIALIZABLE_DERIVED_EVENT(
        dclgcf::message::EnqueueNDRangeKernel,
        dclgcf::message::Request,
        ADD_ATTRIBUTE_DEFAULT(dclgcf::message::EnqueueNDRangeKernel, _commandQueueId);
        ADD_ATTRIBUTE_DEFAULT(dclgcf::message::EnqueueNDRangeKernel, _commandId);
        ADD_ATTRIBUTE_DEFAULT(dclgcf::message::EnqueueNDRangeKernel, _kernelId);
        ADD_ATTRIBUTE_CUSTOM(
                dclgcf::message::EnqueueNDRangeKernel,
                _offset,
                gcf::ST_Custom1,
                std::vector<gcf_uint32>,
                256,
                gcf::Reliable,
                gcf::Public);
        ADD_ATTRIBUTE_CUSTOM(
                dclgcf::message::EnqueueNDRangeKernel,
                _global,
                gcf::ST_Custom1,
                std::vector<gcf_uint32>,
                256,
                gcf::Reliable,
                gcf::Public);
        ADD_ATTRIBUTE_CUSTOM(
                dclgcf::message::EnqueueNDRangeKernel,
                _local,
                gcf::ST_Custom1,
                std::vector<gcf_uint32>,
                256,
                gcf::Reliable,
                gcf::Public);
        ADD_ATTRIBUTE_DEFAULT(dclgcf::message::EnqueueNDRangeKernel, _eventIdWaitList);
        ADD_ATTRIBUTE_DEFAULT(dclgcf::message::EnqueueNDRangeKernel, _event);
        )
