/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/
/**
 * @file SetKernelArg.cpp
 *
 * @date 2011-08-21
 * @author Philipp Kegel
 */

#include <dclgcf/message/SetKernelArg.h>
#include <dclgcf/message/Request.h>

#include <dcl/Binary.h>
#include <dcl/DCLTypes.h>

#include <gcf/GCFTypedefs.h>
#include <gcf/GenericSerializerImplNew.cpp>
#include <gcf/Serializable.h>

#include <gcf/serialization/ByteBuffer.h>

#ifdef __APPLE__
#include <OpenCL/cl.h>
#else
#include <CL/cl.h>
#endif

#include <cassert>
#include <cstddef>

/******************************************************************************
 * Set kernel argument
 ******************************************************************************/

namespace dclgcf {

namespace message {

SetKernelArg::SetKernelArg() :
	_kernelId(0) {
	initSerializable();
}

SetKernelArg::SetKernelArg(
		dcl::object_id kernelId,
		cl_uint index) :
	_kernelId(kernelId), _index(static_cast<gcf_uint32>(index)) {
	initSerializable();
}

SetKernelArg::SetKernelArg(
		const SetKernelArg& rhs) :
	Request(rhs), _kernelId(rhs._kernelId), _index(rhs._index) {
	initSerializable();
}

SetKernelArg::~SetKernelArg() { }

dcl::object_id SetKernelArg::kernelId() const {
	return _kernelId;
}

cl_uint SetKernelArg::argIndex() const {
	return static_cast<cl_uint>(_index);
}

} /* namespace message */

} /* namespace dclgcf */

IMPLEMENT_SERIALIZABLE_DERIVED_EVENT(
		dclgcf::message::SetKernelArg,
		dclgcf::message::Request,
		ADD_ATTRIBUTE_DEFAULT(dclgcf::message::SetKernelArg, _kernelId);
		ADD_ATTRIBUTE_DEFAULT(dclgcf::message::SetKernelArg, _index);
		)

/******************************************************************************
 * Set binary kernel argument
 ******************************************************************************/

namespace gcf {

template<>
size_t writeCustomSTR<dclgcf::message::SetKernelArgBinary> (
        dclgcf::message::SetKernelArgBinary *object,
        const gcf::GenericSerializeTemplateRecord& gstr,
        char *buffer) {
    assert(gstr.getType() == gcf::ST_Custom1);

    gcf::ByteBuffer buf(buffer, gstr.getLength());
    dcl::Binary& argValue =
            *gcf::GenericSerializeTemplateRecord::getAttribute<
                    dclgcf::message::SetKernelArgBinary, dcl::Binary,
                    gcf::ST_Custom1>(object, gstr);

    buf << argValue;

    return buf.getLength();
}

template<>
size_t readCustomSTR<dclgcf::message::SetKernelArgBinary> (
        dclgcf::message::SetKernelArgBinary *object,
        const gcf::GenericSerializeTemplateRecord& gstr,
        const char *buffer) {
    assert(gstr.getType() == gcf::ST_Custom1);

    gcf::ByteBuffer buf(buffer, gstr.getLength());
    dcl::Binary& argValue =
            *gcf::GenericSerializeTemplateRecord::getAttribute<
                    dclgcf::message::SetKernelArgBinary, dcl::Binary,
                    gcf::ST_Custom1>(object, gstr);

    buf >> argValue;

    return buf.getLength();
}

} /* namespace gcf */

/******************************************************************************/

namespace dclgcf {

namespace message {

SetKernelArgBinary::SetKernelArgBinary() {
	initSerializable();
}

SetKernelArgBinary::SetKernelArgBinary(
		dcl::object_id kernelId,
		cl_uint index,
		size_t size,
		const void *value) :
	SetKernelArg(kernelId, index), _arg(size, value) {
	initSerializable();
}

SetKernelArgBinary::SetKernelArgBinary(
		const SetKernelArgBinary& rhs) :
	SetKernelArg(rhs), _arg(rhs._arg) {
	initSerializable();
}

SetKernelArgBinary::~SetKernelArgBinary() { }

size_t SetKernelArgBinary::argSize() const {
	return _arg.size();
}

const void * SetKernelArgBinary::argValue() const {
	return _arg.value();
}

} /* namespace message */

} /* namespace dclgcf */

IMPLEMENT_SERIALIZABLE_DERIVED_EVENT(
		dclgcf::message::SetKernelArgBinary,
		dclgcf::message::SetKernelArg,
		ADD_ATTRIBUTE_CUSTOM(dclgcf::message::SetKernelArgBinary, _arg,
				gcf::ST_Custom1, dcl::Binary, 256, gcf::Reliable,
				gcf::Public);
		)

/******************************************************************************
 * Set memory object as kernel argument
 ******************************************************************************/

namespace dclgcf {

namespace message {

SetKernelArgMemObject::SetKernelArgMemObject() :
	_memObjectId(0) {
	initSerializable();
}

SetKernelArgMemObject::SetKernelArgMemObject(
		dcl::object_id kernelId,
		cl_uint index,
		size_t size) :
	SetKernelArg(kernelId, index), _size(static_cast<gcf_uint32>(size)),
	_memObjectId(0) {
	initSerializable();
}

SetKernelArgMemObject::SetKernelArgMemObject(
		dcl::object_id kernelId,
		cl_uint index,
		dcl::object_id memObjectId) :
	SetKernelArg(kernelId, index), _size(static_cast<gcf_uint32>(sizeof(cl_mem))),
	_memObjectId(memObjectId) {
	initSerializable();
}

SetKernelArgMemObject::SetKernelArgMemObject(
		const SetKernelArgMemObject& rhs) :
	SetKernelArg(rhs), _size(rhs._size), _memObjectId(rhs._memObjectId) {
	initSerializable();
}

SetKernelArgMemObject::~SetKernelArgMemObject() { }

size_t SetKernelArgMemObject::argSize() const {
	return static_cast<size_t>(_size);
}

const void * SetKernelArgMemObject::argValue() const {
	return (_memObjectId == 0) ? nullptr : &_memObjectId;
}

dcl::object_id SetKernelArgMemObject::memObjectId() const {
	return _memObjectId;
}

} /* namespace message */

} /* namespace dclgcf */

IMPLEMENT_SERIALIZABLE_DERIVED_EVENT(
        dclgcf::message::SetKernelArgMemObject,
        dclgcf::message::SetKernelArg,
        ADD_ATTRIBUTE_DEFAULT(dclgcf::message::SetKernelArgMemObject, _size);
        ADD_ATTRIBUTE_DEFAULT(dclgcf::message::SetKernelArgMemObject, _memObjectId);
        )
