/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/
/**
 * @file Buffer.cpp
 *
 * @date 2013-04-07
 * @author Philipp Kegel
 */

#include <cstddef>
#include <cstring>
#include <iostream>

#ifdef __APPLE__
#include <OpenCL/cl.h>
#include <OpenCL/cl_wwu_dcl.h>
#else
#include <CL/cl.h>
#include <CL/cl_wwu_dcl.h>
#endif

#define BOOST_TEST_DYN_LINK
#define BOOST_TEST_MODULE Memory
#include <boost/test/unit_test.hpp>

/**
 * @brief Returns the first platform available
 *
 * @return a platform
 */
cl_platform_id getPlatform() {
    cl_platform_id platform;
    cl_uint num_platforms;
    size_t param_value_size;
    cl_int err;

    err = clGetPlatformIDs(1, &platform, &num_platforms);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    BOOST_REQUIRE_MESSAGE(num_platforms >= 1, "No platform found");

    /* platform should be 'dOpenCL' */
    err = clGetPlatformInfo(platform, CL_PLATFORM_NAME, 0, NULL,
            &param_value_size);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    char platform_name[param_value_size];
    err = clGetPlatformInfo(platform, CL_PLATFORM_NAME, param_value_size,
            platform_name, NULL);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    BOOST_CHECK_MESSAGE(strcmp(platform_name, "dOpenCL") == 0, "First platform must be 'dOpenCL'");

    return platform;
}

/**
 * @brief Obtains the specified number of compute nodes from the given platform.
 *
 * @param platform
 * @param num_nodes
 * @param nodes
 */
void getComputeNodes(cl_platform_id platform, cl_uint num_nodes, cl_compute_node_WWU *nodes) {
    cl_uint num_nodes_ret;
    cl_int err;

    err = clGetComputeNodesWWU(platform, num_nodes, nodes, &num_nodes_ret);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    BOOST_REQUIRE_MESSAGE(num_nodes_ret >= num_nodes, "No enough compute nodes");
}

/**
 * @brief Obtains the specified number of compute nodes from the given platform.
 *
 * @param[in]  platform the platform
 */
void getDevices(cl_platform_id platform, cl_device_type type, cl_uint num_devices, cl_device_id *devices) {
    cl_uint num_devices_ret;
    cl_int err;

    err = clGetDeviceIDs(platform, type, num_devices, devices,
            &num_devices_ret);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    BOOST_REQUIRE_MESSAGE(num_devices_ret >= num_devices, "No enough devices");
}

/******************************************************************************
 * Test cases
 ******************************************************************************/

BOOST_AUTO_TEST_CASE( CreateContext )
{
    cl_platform_id platform = getPlatform();
    cl_device_id device;
    cl_context context;
    cl_int err;

    context = clCreateContext(0, 1, 0, 0, 0, &err);
    BOOST_REQUIRE_EQUAL(err, CL_INVALID_VALUE);
    context = clCreateContext(0, 0, &device, 0, 0, &err);
    BOOST_REQUIRE_EQUAL(err, CL_INVALID_VALUE);
    context = clCreateContext(0, 1, &device, 0, &err, &err);
    BOOST_REQUIRE_EQUAL(err, CL_INVALID_VALUE);

    getDevices(platform, CL_DEVICE_TYPE_ALL, 1, &device);

    /* create context */
    context = clCreateContext(0, 1, &device, 0, 0, &err);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);

    /* clean up */
    err = clReleaseContext(context);
    BOOST_CHECK_EQUAL(err, CL_SUCCESS);
}

BOOST_AUTO_TEST_CASE( CreateContextFromComputeNode )
{
    cl_platform_id platform = getPlatform();
    cl_compute_node_WWU compute_node;
    cl_context context;
    cl_int err;

    context = clCreateContextFromComputeNodesWWU(0, 1, 0, 0, 0, &err);
    BOOST_REQUIRE_EQUAL(err, CL_INVALID_VALUE);
    context = clCreateContextFromComputeNodesWWU(0, 0, &compute_node, 0, 0, &err);
    BOOST_REQUIRE_EQUAL(err, CL_INVALID_VALUE);
    context = clCreateContextFromComputeNodesWWU(0, 0, &compute_node, 0, &err, &err);
    BOOST_REQUIRE_EQUAL(err, CL_INVALID_VALUE);

    getComputeNodes(platform, 1, &compute_node);

    /* create context from compute node */
    context = clCreateContextFromComputeNodesWWU(0, 1, &compute_node, 0, 0, &err);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);

    /* verify context */
    cl_uint num_devices, num_devices_ctx;

    err = clGetDeviceIDsFromComputeNodeWWU(compute_node, CL_DEVICE_TYPE_ALL, 0, 0, &num_devices);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    err = clGetContextInfo(context, CL_CONTEXT_NUM_DEVICES, sizeof(num_devices), &num_devices_ctx, 0);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    BOOST_REQUIRE_EQUAL(num_devices_ctx, num_devices);

    cl_device_id devices[num_devices];
    cl_device_id devices_ctx[num_devices_ctx];

    err = clGetDeviceIDsFromComputeNodeWWU(compute_node, CL_DEVICE_TYPE_ALL, num_devices, devices, 0);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    err = clGetContextInfo(context, CL_CONTEXT_DEVICES, sizeof(devices), &devices_ctx, 0);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    for (int i = 0; i < num_devices; ++i) {
        BOOST_REQUIRE_EQUAL(devices_ctx[i], devices[i]);
    }

    /* clean up */
    err = clReleaseContext(context);
    BOOST_CHECK_EQUAL(err, CL_SUCCESS);
}

BOOST_AUTO_TEST_CASE( CreateContextFromType )
{
    cl_platform_id platform = getPlatform();
    cl_context context;
    cl_int err;

    context = clCreateContextFromType(0, CL_DEVICE_TYPE_ALL, 0, &err, &err);
    BOOST_REQUIRE_EQUAL(err, CL_INVALID_VALUE);

    /* create context from type */
    context = clCreateContextFromType(0, CL_DEVICE_TYPE_ALL, 0, 0, &err);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);

    /* clean up */
    err = clReleaseContext(context);
    BOOST_CHECK_EQUAL(err, CL_SUCCESS);
}
