/******************************************************************************
 * This file is part of dOpenCL.
 * 
 * dOpenCL is an implementation of the OpenCL application programming
 * interface for distributed systems. See <http://dopencl.uni-muenster.de/>
 * for more information.
 * 
 * Developed by: Research Group Parallel and Distributed Systems
 *               Department of Mathematics and Computer Science
 *               University of Muenster, Germany
 *               <http://pvs.uni-muenster.de/>
 * 
 * Copyright (C) 2013  Philipp Kegel <philipp.kegel@uni-muenster.de>
 *
 * dOpenCL is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * dOpenCL is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with dOpenCL. If not, see <http://www.gnu.org/licenses/>.
 * 
 * Permission to use dOpenCL for scientific, non-commercial work is
 * granted under the terms of the dOpenCL Academic License provided
 * appropriate credit is given. See the dOpenCL Academic License for
 * more details.
 * 
 * You should have received a copy of the dOpenCL Academic License
 * along with dOpenCL. If not, see <http://dopencl.uni-muenster.de/>.
 ******************************************************************************/
/**
 * @file MemoryConsistency.cpp
 *
 * @date 2013-02-18
 * @author Philipp Kegel
 */

#include <cstddef>
#include <cstring>
#include <iostream>

#ifdef __APPLE__
#include <OpenCL/cl.h>
#include <OpenCL/cl_wwu_dcl.h>
#else
#include <CL/cl.h>
#include <CL/cl_wwu_dcl.h>
#endif

#define BOOST_TEST_DYN_LINK
#define BOOST_TEST_MODULE MemoryConsistency
#include <boost/test/unit_test.hpp>

/**
 * @brief Returns the first platform available
 *
 * @return a platform
 */
cl_platform_id getPlatform() {
    cl_platform_id platform;
    cl_uint num_platforms;
    size_t param_value_size;
    cl_int err;

    err = clGetPlatformIDs(1, &platform, &num_platforms);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    BOOST_REQUIRE_MESSAGE(num_platforms >= 1, "No platform found");

    /* platform should be 'dOpenCL' */
    err = clGetPlatformInfo(platform, CL_PLATFORM_NAME, 0, NULL,
            &param_value_size);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    char platform_name[param_value_size];
    err = clGetPlatformInfo(platform, CL_PLATFORM_NAME, param_value_size,
            platform_name, NULL);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    BOOST_CHECK_MESSAGE(strcmp(platform_name, "dOpenCL") == 0, "First platform must be 'dOpenCL'");

    return platform;
}

/**
 * @brief Obtains the specified number of compute nodes from the given platform.
 *
 * @param platform
 * @param num_nodes
 * @param nodes
 */
void getComputeNodes(cl_platform_id platform, cl_uint num_nodes, cl_compute_node_WWU *nodes) {
    cl_uint num_nodes_ret;
    cl_int err;

    err = clGetComputeNodesWWU(platform, num_nodes, nodes, &num_nodes_ret);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    BOOST_REQUIRE_MESSAGE(num_nodes_ret >= num_nodes, "No enough compute nodes");
}

/**
 * @brief Returns the compute node's first device.
 *
 * @param[in]  node the compute node
 * @return the compute node's first device
 */
cl_device_id getDevice(cl_compute_node_WWU node) {
    cl_device_id device = NULL;
    cl_uint num_devices;
    cl_int err;

    err = clGetDeviceIDsFromComputeNodeWWU(node, CL_DEVICE_TYPE_ALL, 1, &device,
            &num_devices);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    BOOST_REQUIRE_GE(num_devices, 1);

    return device;
}

cl_context createContext(cl_uint num_devices, cl_device_id *devices) {
    cl_context context;
    cl_int err;

    BOOST_REQUIRE_GE(num_devices, 2);
    BOOST_REQUIRE(devices != NULL);

    context = clCreateContext(0, num_devices, devices, 0, 0, &err);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);

    return context;
}

cl_command_queue createCommandQueue(cl_context context, cl_device_id device) {
    cl_command_queue commandQueue;
    cl_int err;

    commandQueue = clCreateCommandQueue(context, device, 0, &err);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);

    return commandQueue;
}

cl_mem createRWBuffer(cl_context context, size_t cb) {
    cl_mem buffer = 0;
    cl_int err;

    buffer = clCreateBuffer(context, CL_MEM_READ_WRITE, cb, NULL, &err);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);

    return buffer;
}

/******************************************************************************
 * Shared test suite fixture
 ******************************************************************************/

struct Context {
    cl_device_id devices[2];
    cl_command_queue commandQueues[2];
    cl_context context;
    cl_mem buffer;
    size_t vecSize;
    size_t cb;

    Context() :
            vecSize(1024 * 1024) /* #vector elements */,
            cb(vecSize * sizeof(cl_int)) /* buffer size */
    {
        cl_platform_id platform = getPlatform();
        cl_compute_node_WWU nodes[2];

        getComputeNodes(platform, 2, nodes);
        devices[0] = getDevice(nodes[0]);
        devices[1] = getDevice(nodes[1]);

        context = createContext(2, devices);
        commandQueues[0] = createCommandQueue(context, devices[0]);
        commandQueues[1] = createCommandQueue(context, devices[1]);
        buffer = createRWBuffer(context, cb);

        BOOST_TEST_MESSAGE("Setup fixture");
    }

    ~Context() {
        cl_int err;

        /* clean up */
        err = clReleaseMemObject(buffer);
        BOOST_CHECK_EQUAL(err, CL_SUCCESS);
        err = clReleaseCommandQueue(commandQueues[0]);
        BOOST_CHECK_EQUAL(err, CL_SUCCESS);
        err = clReleaseCommandQueue(commandQueues[1]);
        BOOST_CHECK_EQUAL(err, CL_SUCCESS);
        err = clReleaseContext(context);
        BOOST_CHECK_EQUAL(err, CL_SUCCESS);

        BOOST_TEST_MESSAGE("Teared down fixture");
    }
};

BOOST_FIXTURE_TEST_SUITE( MemoryConsistency, Context )

/******************************************************************************
 * Test cases
 ******************************************************************************/

template<typename T>
void fillVector(std::vector<T>& vec, T first, T step) {
    T value = first;
    for (typename std::vector<T>::iterator i = vec.begin(); i != vec.end(); ++i) {
        *i = value;
        value += step;
    }
}

BOOST_AUTO_TEST_CASE( WriteRead )
{
    cl_event write;
    std::vector<cl_int> vecIn(vecSize, 0), vecOut(vecSize, 1);
    cl_int err;

    fillVector(vecIn, 1, 1); // initialize input data

    /* upload data to device on first compute node */
    err = clEnqueueWriteBuffer(commandQueues[0], buffer, CL_FALSE, 0, cb,
            &vecIn.front(), 0, NULL, &write);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    /* ensure start of data upload to first compute node */
    err = clFlush(commandQueues[0]);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    /* download data from device on second compute node */
    err = clEnqueueReadBuffer(commandQueues[1], buffer, CL_TRUE, 0, cb,
            &vecOut.front(), 1, &write, NULL);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);

    BOOST_CHECK_MESSAGE(vecIn == vecOut, "Input and output buffers differ"); // compare input and output data

    /* clean up */
    err = clReleaseEvent(write);
    BOOST_CHECK_EQUAL(err, CL_SUCCESS);
}

BOOST_AUTO_TEST_CASE( NDRangeKernelRead )
{
    const char *source = "\
__kernel void init(__global int *v) {       \
    v[get_global_id(0)] = get_global_id(0); \
}";
    cl_program program;
    cl_kernel kernel;
    cl_event init;
    std::vector<cl_int> hVec(vecSize, 0), dVec(vecSize, 1);
    cl_int err;

    fillVector(hVec, 0, 1); // initialize host vector

    program = clCreateProgramWithSource(context, 1, &source, NULL, &err);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    err = clBuildProgram(program, 2, devices, NULL, NULL, NULL);
    kernel = clCreateKernel(program, "init", &err);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    err = clSetKernelArg(kernel, 0, sizeof(cl_mem), &buffer);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);

    /* submit kernel to device on first compute node */
    err = clEnqueueNDRangeKernel(commandQueues[0], kernel,
            1, NULL, &vecSize, NULL, 0, NULL, &init);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    /* ensure start of kernel execution on first compute node */
    err = clFlush(commandQueues[0]);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    /* download data from device on second compute node */
    err = clEnqueueReadBuffer(commandQueues[1], buffer, CL_TRUE, 0, cb,
            &dVec.front(), 1, &init, NULL);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);

    BOOST_CHECK_MESSAGE(hVec == dVec, "Host and device buffers differ"); // compare host and device buffer

    /* clean up */
    err = clReleaseEvent(init);
    BOOST_CHECK_EQUAL(err, CL_SUCCESS);
    err = clReleaseKernel(kernel);
    BOOST_CHECK_EQUAL(err, CL_SUCCESS);
    err = clReleaseProgram(program);
    BOOST_CHECK_EQUAL(err, CL_SUCCESS);
}

/* Test cross-over exchange of two memory objects */
BOOST_AUTO_TEST_CASE( ConcurrentWriteRead )
{
    cl_mem buffer1;
    cl_event write[2], read[2];
    std::vector<cl_int> vecIn0(vecSize, 1), vecIn1(vecSize, 2);
    std::vector<cl_int> vecOut0(vecSize, 0), vecOut1(vecSize, 0);
    cl_int err;

    buffer1 = createRWBuffer(context, cb);

    /* upload data to device on first compute node */
    err = clEnqueueWriteBuffer(commandQueues[0], buffer, CL_FALSE, 0, cb,
            &vecIn0.front(), 0, NULL, &write[0]);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    /* upload data to device on second compute node */
    err = clEnqueueWriteBuffer(commandQueues[1], buffer1, CL_FALSE, 0, cb,
            &vecIn1.front(), 0, NULL, &write[1]);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);

    /* download data from device on first compute node */
    err = clEnqueueReadBuffer(commandQueues[0], buffer1, CL_FALSE, 0, cb,
            &vecOut0.front(), 1, &write[1], &read[0]);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);
    /* download data from device on second compute node */
    err = clEnqueueReadBuffer(commandQueues[1], buffer, CL_FALSE, 0, cb,
            &vecOut1.front(), 1, &write[0], &read[1]);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);

    /* await completion of data transfers */
    err = clWaitForEvents(2, read);
    BOOST_REQUIRE_EQUAL(err, CL_SUCCESS);

    BOOST_CHECK_MESSAGE(vecIn0 == vecOut1, "Input and output buffers differ"); // compare input and output data
    BOOST_CHECK_MESSAGE(vecIn1 == vecOut0, "Input and output buffers differ"); // compare input and output data

    /* clean up */
    err = clReleaseEvent(write[0]);
    BOOST_CHECK_EQUAL(err, CL_SUCCESS);
    err = clReleaseEvent(write[1]);
    BOOST_CHECK_EQUAL(err, CL_SUCCESS);
    err = clReleaseEvent(read[0]);
    BOOST_CHECK_EQUAL(err, CL_SUCCESS);
    err = clReleaseEvent(read[1]);
    BOOST_CHECK_EQUAL(err, CL_SUCCESS);
    err = clReleaseMemObject(buffer1);
    BOOST_CHECK_EQUAL(err, CL_SUCCESS);
}

BOOST_AUTO_TEST_SUITE_END()
